import asyncio
from collections.abc import Iterable, Sequence
from enum import Enum
from pathlib import Path
from typing import Optional

from anyio import Path as APath
from hishel import AsyncFileStorage
import humanize
import inflect
from rich.console import Console, ConsoleRenderable
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from uv_secure.configuration import (
    config_cli_arg_factory,
    config_file_factory,
    Configuration,
    override_config,
)
from uv_secure.directory_scanner import get_dependency_file_to_config_map
from uv_secure.package_info import (
    download_packages,
    PackageInfo,
    parse_requirements_txt_file,
    parse_uv_lock_file,
)


def _render_vulnerability_table(
    config: Configuration, vulnerable_packages: Iterable[PackageInfo]
) -> Table:
    table = Table(
        title="Vulnerable Dependencies",
        show_header=True,
        row_styles=["none", "dim"],
        header_style="bold magenta",
        expand=True,
    )
    table.add_column("Package", min_width=8, max_width=40)
    table.add_column("Version", min_width=10, max_width=20)
    table.add_column("Vulnerability ID", style="bold cyan", min_width=20, max_width=24)
    table.add_column("Fix Versions", min_width=10, max_width=20)
    if config.aliases:
        table.add_column("Aliases", min_width=20, max_width=24)
    if config.desc:
        table.add_column("Details", min_width=8)
    for package in vulnerable_packages:
        for vuln in package.vulnerabilities:
            vuln_id_hyperlink = (
                Text.assemble((vuln.id, f"link {vuln.link}"))
                if vuln.link
                else Text(vuln.id)
            )
            renderables = [
                Text.assemble(
                    (
                        package.info.name,
                        f"link https://pypi.org/project/{package.info.name}",
                    )
                ),
                Text.assemble(
                    (
                        package.info.version,
                        f"link https://pypi.org/project/{package.info.name}/"
                        f"{package.info.version}/",
                    )
                ),
                vuln_id_hyperlink,
                Text(", ").join(
                    [
                        Text.assemble(
                            (
                                fix_ver,
                                f"link https://pypi.org/project/{package.info.name}/"
                                f"{fix_ver}/",
                            )
                        )
                        for fix_ver in vuln.fixed_in
                    ]
                )
                if vuln.fixed_in
                else Text(""),
            ]
            if config.aliases:
                alias_links = []
                for alias in vuln.aliases or []:
                    hyperlink = None
                    if alias.startswith("CVE-"):
                        hyperlink = (
                            f"https://cve.mitre.org/cgi-bin/cvename.cgi?name={alias}"
                        )
                    elif alias.startswith("GHSA-"):
                        hyperlink = f"https://github.com/advisories/{alias}"
                    elif alias.startswith("PYSEC-"):
                        hyperlink = (
                            "https://github.com/pypa/advisory-database/blob/main/"
                            f"vulns/{package.info.name}/{alias}.yaml"
                        )
                    elif alias.startswith("OSV-"):
                        hyperlink = f"https://osv.dev/vulnerability/{alias}"
                    if hyperlink:
                        alias_links.append(Text.assemble((alias, f"link {hyperlink}")))
                    else:
                        alias_links.append(Text(alias))
                renderables.append(
                    Text(", ").join(alias_links) if alias_links else Text("")
                )
            if config.desc:
                renderables.append(vuln.details)
            table.add_row(*renderables)
    return table


def _render_issue_table(
    config: Configuration, maintenance_issue_packages: Iterable[PackageInfo]
) -> Table:
    table = Table(
        title="Maintenance Issues",
        show_header=True,
        row_styles=["none", "dim"],
        header_style="bold magenta",
        expand=True,
    )
    table.add_column("Package", min_width=8, max_width=40)
    table.add_column("Version", min_width=10, max_width=20)
    table.add_column("Yanked", style="bold cyan", min_width=10, max_width=10)
    table.add_column("Yanked Reason", min_width=20, max_width=24)
    table.add_column("Age", min_width=20, max_width=24)
    for package in maintenance_issue_packages:
        renderables = [
            Text.assemble(
                (
                    package.info.name,
                    f"link https://pypi.org/project/{package.info.name}",
                )
            ),
            Text.assemble(
                (
                    package.info.version,
                    f"link https://pypi.org/project/{package.info.name}/"
                    f"{package.info.version}/",
                )
            ),
            str(package.info.yanked),
            package.info.yanked_reason if package.info.yanked_reason else "Unknown",
            humanize.precisedelta(package.age, minimum_unit="days")
            if package.age
            else "Unknown",
        ]
        table.add_row(*renderables)
    return table


async def check_dependencies(
    dependency_file_path: APath, config: Configuration
) -> tuple[int, Iterable[ConsoleRenderable]]:
    """Checks dependencies for vulnerabilities and summarizes the results

    Args:
        dependency_file_path: uv.lock file or requirements.txt file path
        config: uv-secure configuration object

    Returns:
        tuple with status code and output for console to render
    """
    console_outputs = []

    if not await dependency_file_path.exists():
        console_outputs.append(
            f"[bold red]Error:[/] File {dependency_file_path} does not exist."
        )
        return 3, console_outputs

    # I found antivirus programs (specifically Windows Defender) can almost fully
    # negate the benefits of using a file cache if you don't exclude the virus checker
    # from checking the cache dir given it is frequently read from
    storage = AsyncFileStorage(
        base_path=config.cache_settings.cache_path,
        ttl=config.cache_settings.ttl_seconds,
    )

    if dependency_file_path.name == "uv.lock":
        dependencies = await parse_uv_lock_file(dependency_file_path)
    else:  # Assume dependency_file_path.name == "requirements.txt"
        dependencies = await parse_requirements_txt_file(dependency_file_path)

    if len(dependencies) == 0:
        return 0, console_outputs

    console_outputs.append(
        f"[bold cyan]Checking {dependency_file_path} dependencies for vulnerabilities"
        "...[/]\n"
    )

    packages = await download_packages(
        dependencies, storage, config.cache_settings.disable_cache
    )

    total_dependencies = len(packages)
    vulnerable_count = 0
    vulnerable_packages = []
    maintenance_issue_packages = []

    for idx, package in enumerate(packages):
        if isinstance(package, BaseException):
            console_outputs.append(
                f"[bold red]Error:[/] {dependencies[idx]} raised exception: {package}"
            )
            continue

        # Filter out ignored vulnerabilities
        package.vulnerabilities = [
            vuln
            for vuln in package.vulnerabilities
            if config.ignore_vulnerabilities is None
            or vuln.id not in config.ignore_vulnerabilities
        ]
        if len(package.vulnerabilities) > 0:
            vulnerable_count += len(package.vulnerabilities)
            vulnerable_packages.append(package)

        found_rejected_yanked_package = (
            config.maintainability_criteria.forbid_yanked and package.info.yanked
        )
        found_over_age_package = (
            config.maintainability_criteria.max_package_age is not None
            and package.age is not None
            and package.age > config.maintainability_criteria.max_package_age
        )
        if found_rejected_yanked_package or found_over_age_package:
            maintenance_issue_packages.append(package)

    inf = inflect.engine()
    total_plural = inf.plural("dependency", total_dependencies)
    vulnerable_plural = inf.plural("vulnerability", vulnerable_count)

    status = 0
    if vulnerable_count > 0:
        console_outputs.append(
            Panel.fit(
                f"[bold red]Vulnerabilities detected![/]\n"
                f"Checked: [bold]{total_dependencies}[/] {total_plural}\n"
                f"Vulnerable: [bold]{vulnerable_count}[/] {vulnerable_plural}"
            )
        )
        table = _render_vulnerability_table(config, vulnerable_packages)
        console_outputs.append(table)
        status = 2

    issue_count = len(maintenance_issue_packages)
    issue_plural = inf.plural("issue", issue_count)
    if len(maintenance_issue_packages) > 0:
        console_outputs.append(
            Panel.fit(
                f"[bold yellow]Maintenance Issues detected![/]\n"
                f"Checked: [bold]{total_dependencies}[/] {total_plural}\n"
                f"Issues: [bold]{issue_count}[/] {issue_plural}"
            )
        )
        table = _render_issue_table(config, maintenance_issue_packages)
        console_outputs.append(table)
        status = max(status, 1)

    if status == 0:
        console_outputs.append(
            Panel.fit(
                f"[bold green]No vulnerabilities detected![/]\n"
                f"Checked: [bold]{total_dependencies}[/] {total_plural}\n"
                f"All dependencies appear safe!"
            )
        )
    return status, console_outputs


class RunStatus(Enum):
    NO_VULNERABILITIES = (0,)
    MAINTENANCE_ISSUES_FOUND = 1
    VULNERABILITIES_FOUND = 2
    RUNTIME_ERROR = 3


async def check_lock_files(
    file_paths: Optional[Sequence[Path]],
    aliases: Optional[bool],
    desc: Optional[bool],
    disable_cache: Optional[bool],
    forbid_yanked: Optional[bool],
    max_package_age: Optional[int],
    ignore: Optional[str],
    config_path: Optional[Path],
) -> RunStatus:
    """Checks

    Args:
        file_paths: paths to files or directory to process
        aliases: flag whether to show vulnerability aliases
        desc: flag whether to show vulnerability descriptions
        disable_cache: flag whether to disable cache
        forbid_yanked: flag whether to forbid yanked dependencies
        max_package_age: maximum age of dependencies in days
        ignore: Vulnerabilities IDs to ignore

    Returns
    -------
        True if vulnerabilities were found, False otherwise.
    """
    if not file_paths:
        file_paths = (Path(),)

    console = Console()
    if len(file_paths) == 1 and file_paths[0].is_dir():
        lock_to_config_map = await get_dependency_file_to_config_map(
            APath(file_paths[0])
        )
        file_paths = tuple(lock_to_config_map.keys())
    else:
        if config_path is not None:
            possible_config = await config_file_factory(APath(config_path))
            config = possible_config if possible_config is not None else Configuration()
            lock_to_config_map = {APath(file): config for file in file_paths}
        elif all(
            file_path.name in {"requirements.txt", "uv.lock"}
            for file_path in file_paths
        ):
            lock_to_config_map = await get_dependency_file_to_config_map(
                [APath(file_path) for file_path in file_paths]
            )
            file_paths = tuple(lock_to_config_map.keys())
        else:
            console.print(
                "[bold red]Error:[/] file_paths must either reference a single "
                "project root directory or a sequence of uv.lock / requirements.txt "
                "file paths"
            )
            return RunStatus.RUNTIME_ERROR

    if any(
        (
            aliases,
            desc,
            ignore,
            disable_cache,
            forbid_yanked,
            max_package_age is not None,
        )
    ):
        cli_config = config_cli_arg_factory(
            aliases, desc, disable_cache, forbid_yanked, max_package_age, ignore
        )
        lock_to_config_map = {
            lock_file: override_config(config, cli_config)
            for lock_file, config in lock_to_config_map.items()
        }

    status_output_tasks = [
        check_dependencies(
            APath(dependency_file_path), lock_to_config_map[APath(dependency_file_path)]
        )
        for dependency_file_path in file_paths
    ]
    status_outputs = await asyncio.gather(*status_output_tasks)
    maintenance_issues_found = False
    vulnerabilities_found = False
    runtime_error = False
    for status, console_output in status_outputs:
        console.print(*console_output)
        if status == 1:
            maintenance_issues_found = True
        elif status == 2:
            vulnerabilities_found = True
        elif status == 3:
            runtime_error = True
    if runtime_error:
        return RunStatus.RUNTIME_ERROR
    if vulnerabilities_found:
        return RunStatus.VULNERABILITIES_FOUND
    if maintenance_issues_found:
        return RunStatus.MAINTENANCE_ISSUES_FOUND
    return RunStatus.NO_VULNERABILITIES



from pathlib import Path
from textwrap import dedent

import pytest


@pytest.fixture
def temp_uv_lock_file(tmp_path: Path) -> Path:
    """Fixture to create a temporary uv.lock file with a single dependency."""
    uv_lock_path = tmp_path / "uv.lock"
    uv_lock_data = """
        [[package]]
        name = "example-package"
        version = "1.0.0"
        source = { registry = "https://pypi.org/simple" }
    """
    uv_lock_path.write_text(dedent(uv_lock_data).strip())
    return uv_lock_path



from uv_secure.package_info.dependency_file_parser import (
    parse_requirements_txt_file,
    parse_uv_lock_file,
)
from uv_secure.package_info.package_info_downloader import (
    download_packages,
    PackageInfo,
)


__all__ = [
    "PackageInfo",
    "download_packages",
    "parse_requirements_txt_file",
    "parse_uv_lock_file",
]



from datetime import timedelta
import sys
from typing import Optional

from anyio import Path

from uv_secure.configuration.configuration import Configuration, OverrideConfiguration


if sys.version_info >= (3, 11):
    import tomllib as toml
else:
    import tomli as toml


def config_cli_arg_factory(
    aliases: Optional[bool],
    desc: Optional[bool],
    disable_cache: Optional[bool],
    forbid_yanked: Optional[bool],
    max_package_age: Optional[int],
    ignore: Optional[str],
) -> OverrideConfiguration:
    """Factory to create a uv-secure configuration from its command line arguments

    Args:
        aliases: Flag whether to show vulnerability aliases in results
        desc: Flag whether to show vulnerability descriptions in results
        disable_cache: Flag whether to disable cache
        forbid_yanked: flag whether to forbid yanked dependencies
        max_package_age: maximum age of dependencies in days
        ignore: comma separated string of vulnerability ids to ignore

    Returns
    -------
        uv-secure override configuration object
    """
    ignore_vulnerabilities = (
        {vuln_id.strip() for vuln_id in ignore.split(",") if vuln_id.strip()}
        if ignore is not None
        else None
    )

    return OverrideConfiguration(
        aliases=aliases,
        desc=desc,
        disable_cache=disable_cache,
        forbid_yanked=forbid_yanked,
        max_package_age=timedelta(days=max_package_age) if max_package_age else None,
        ignore_vulnerabilities=ignore_vulnerabilities,
    )


async def config_file_factory(config_file: Path) -> Optional[Configuration]:
    """Factory to create a uv-secure configuration from a configuration toml file

    Args:
        config_file: Path to the configuration file (uv-secure.toml, .uv-secure.toml, or
        pyproject.toml)

    Returns
    -------
        uv-secure configuration object or None if no configuration was present
    """
    config_contents = toml.loads(await config_file.read_text())
    if config_file.name == "pyproject.toml":
        if "tool" in config_contents and "uv-secure" in config_contents["tool"]:
            return Configuration(**config_contents["tool"]["uv-secure"])
        return None
    return Configuration(**config_contents)



import asyncio
from datetime import datetime, timedelta
import re
from typing import Optional, Union
from zoneinfo import ZoneInfo

from hishel import AsyncCacheClient, AsyncFileStorage
from pydantic import BaseModel

from uv_secure.package_info.dependency_file_parser import Dependency


class Downloads(BaseModel):
    last_day: Optional[int] = None
    last_month: Optional[int] = None
    last_week: Optional[int] = None


class Info(BaseModel):
    author: Optional[str] = None
    author_email: Optional[str] = None
    bugtrack_url: Optional[str] = None
    classifiers: list[str]
    description: str
    description_content_type: Optional[str] = None
    docs_url: Optional[str] = None
    download_url: Optional[str] = None
    downloads: Downloads
    dynamic: Optional[Union[list[str], str]] = None
    home_page: Optional[str] = None
    keywords: Optional[Union[str, list[str]]] = None
    license: Optional[str] = None
    license_expression: Optional[str] = None
    license_files: Optional[list[str]] = None
    maintainer: Optional[str] = None
    maintainer_email: Optional[str] = None
    name: str
    package_url: Optional[str] = None
    platform: Optional[str] = None
    project_url: Optional[str] = None
    project_urls: Optional[dict[str, str]] = None
    provides_extra: Optional[list[str]] = None
    release_url: str
    requires_dist: Optional[list[str]] = None
    requires_python: Optional[str] = None
    summary: Optional[str] = None
    version: str
    yanked: bool
    yanked_reason: Optional[str] = None


class Digests(BaseModel):
    blake2b_256: str
    md5: str
    sha256: str


class Url(BaseModel):
    comment_text: Optional[str] = None
    digests: Digests
    downloads: int
    filename: str
    has_sig: bool
    md5_digest: str
    packagetype: str
    python_version: str
    requires_python: Optional[str] = None
    size: int
    upload_time: datetime
    upload_time_iso_8601: datetime
    url: str
    yanked: bool
    yanked_reason: Optional[str] = None


class Vulnerability(BaseModel):
    id: str
    details: str
    fixed_in: Optional[list[str]] = None
    aliases: Optional[list[str]] = None
    link: Optional[str] = None
    source: Optional[str] = None
    summary: Optional[str] = None
    withdrawn: Optional[str] = None


class PackageInfo(BaseModel):
    info: Info
    last_serial: int
    urls: list[Url]
    vulnerabilities: list[Vulnerability]

    @property
    def age(self) -> Optional[timedelta]:
        """Return age of the package"""
        release_date = min(
            (url.upload_time_iso_8601 for url in self.urls), default=None
        )
        if release_date is None:
            return None
        return datetime.now(tz=ZoneInfo("UTC")) - release_date


def _canonicalize_name(name: str) -> str:
    """Converts a package name to its canonical form for PyPI URLs"""
    return re.sub(r"[_.]+", "-", name).lower()


async def _download_package(
    client: AsyncCacheClient, dependency: Dependency, disable_cache: bool
) -> PackageInfo:
    """Queries the PyPi JSON API for vulnerabilities of a given dependency."""
    canonical_name = _canonicalize_name(dependency.name)
    url = f"https://pypi.org/pypi/{canonical_name}/{dependency.version}/json"
    response = await client.get(
        url, extensions={"cache_disabled": True} if disable_cache else None
    )
    response.raise_for_status()
    return PackageInfo(**response.json())


async def download_packages(
    dependencies: list[Dependency], storage: AsyncFileStorage, disable_cache: bool
) -> list[Union[PackageInfo, BaseException]]:
    """Fetch vulnerabilities for all dependencies concurrently."""
    async with AsyncCacheClient(timeout=10, storage=storage) as client:
        tasks = [_download_package(client, dep, disable_cache) for dep in dependencies]
        return await asyncio.gather(*tasks, return_exceptions=True)



from pathlib import Path
from textwrap import dedent
from typing import Optional

from anyio import Path as APath
import pytest

from uv_secure.configuration import CacheSettings, config_file_factory, Configuration


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("filename", "file_contents", "expected_configuration"),
    [
        pytest.param(
            "uv-secure.toml",
            "",
            Configuration(),
            id="Empty configuration returns Configuration",
        ),
        pytest.param(
            "uv-secure.toml",
            """
            aliases = true
            desc = true
            """,
            Configuration(aliases=True, desc=True),
            id="Enable aliases and description",
        ),
        pytest.param(
            "uv-secure.toml",
            """
            [cache_settings]
            cache_path = "/dummy/.uv-secure"
            ttl_seconds = 60.0
            """,
            Configuration(
                cache_settings=CacheSettings(
                    cache_path=Path("/dummy/.uv-secure"), ttl_seconds=60.0
                )
            ),
            id="Custom cache settings",
        ),
        pytest.param(
            "pyproject.toml",
            """
            [tool.uv-secure]
            aliases = true
            desc = true
            [tool.uv-secure.cache_settings]
            cache_path = "/dummy/.uv-secure"
            ttl_seconds = 60.0
            """,
            Configuration(
                aliases=True,
                desc=True,
                cache_settings=CacheSettings(
                    cache_path=Path("/dummy/.uv-secure"), ttl_seconds=60.0
                ),
            ),
            id="pyproject.toml aliases, description, and cache_settings",
        ),
        pytest.param(
            "pyproject.toml",
            """
            [tool.some_tool]
            some_option = true
            some_other_option = true
            """,
            None,
            id="pyproject.toml no uv-secure",
        ),
    ],
)
async def test_check_dependencies_alias_hyperlinks(
    tmp_path: Path,
    filename: str,
    file_contents: str,
    expected_configuration: Optional[Configuration],
) -> None:
    config_file_path = tmp_path / filename
    config_file_path.write_text(dedent(file_contents).strip())
    config = await config_file_factory(APath(config_file_path))
    assert config == expected_configuration



import pytest

from uv_secure.configuration import (
    CacheSettings,
    Configuration,
    override_config,
    OverrideConfiguration,
)


@pytest.mark.parametrize(
    ("original", "override", "expected"),
    [
        pytest.param(
            Configuration(aliases=False, desc=False),
            OverrideConfiguration(aliases=True, desc=True),
            Configuration(aliases=True, desc=True),
            id="aliases and desc override to True",
        ),
        pytest.param(
            Configuration(aliases=True, desc=True),
            OverrideConfiguration(aliases=False, desc=False),
            Configuration(aliases=False, desc=False),
            id="aliases and desc override to False",
        ),
        pytest.param(
            Configuration(),
            OverrideConfiguration(disable_cache=True),
            Configuration(cache_settings=CacheSettings(disable_cache=True)),
            id="Disable caching",
        ),
    ],
)
def test_override_config(
    original: Configuration, override: OverrideConfiguration, expected: Configuration
) -> None:
    assert override_config(original, override) == expected



"""Check that basic features work.

Catch cases where e.g. files are missing so the import doesn't work. It is
recommended to check that e.g. assets are included.
"""

import uv_secure


print(uv_secure.__version__)



from uv_secure.configuration.config_factory import (
    config_cli_arg_factory,
    config_file_factory,
)
from uv_secure.configuration.configuration import (
    CacheSettings,
    Configuration,
    override_config,
    OverrideConfiguration,
)


__all__ = [
    "CacheSettings",
    "Configuration",
    "OverrideConfiguration",
    "config_cli_arg_factory",
    "config_file_factory",
    "override_config",
]



from pathlib import Path

from anyio import Path as APath
import pytest
from pytest_httpx import HTTPXMock
from rich.table import Table
from rich.text import Text

from uv_secure.configuration import Configuration
from uv_secure.dependency_checker import check_dependencies


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("alias", "expected_hyperlink"),
    [
        pytest.param(
            "CVE-2024-12345",
            "https://cve.mitre.org/cgi-bin/cvename.cgi?name=CVE-2024-12345",
            id="CVE lias",
        ),
        pytest.param(
            "GHSA-q2x7-8rv6-6q7h",
            "https://github.com/advisories/GHSA-q2x7-8rv6-6q7h",
            id="GHSA alias",
        ),
        pytest.param(
            "PYSEC-12345",
            "https://github.com/pypa/advisory-database/blob/main/vulns/example-package/PYSEC-12345.yaml",
            id="PYSEC alias",
        ),
        pytest.param(
            "OSV-12345", "https://osv.dev/vulnerability/OSV-12345", id="OSV alias"
        ),
        pytest.param("Unrecognised-alias-12345", None, id="Unrecognised alias"),
    ],
)
async def test_check_dependencies_alias_hyperlinks(
    alias: str, expected_hyperlink: str, temp_uv_lock_file: Path, httpx_mock: HTTPXMock
) -> None:
    """Test that aliases generate the correct hyperlink in Rich renderables."""
    # Mock the response to include the alias
    httpx_mock.add_response(
        url="https://pypi.org/pypi/example-package/1.0.0/json",
        json={
            "info": {
                "author_email": "example@example.com",
                "classifiers": [],
                "description": "A minimal package",
                "description_content_type": "text/plain",
                "downloads": {"last_day": None, "last_month": None, "last_week": None},
                "name": "example-package",
                "project_urls": {},
                "provides_extra": [],
                "release_url": "https://pypi.org/project/example-package/1.0.0/",
                "requires_python": ">=3.9",
                "summary": "A minimal package example",
                "version": "1.0.0",
                "yanked": False,
            },
            "last_serial": 1,
            "urls": [],
            "vulnerabilities": [
                {
                    "id": "VULN-123",
                    "details": "Test vulnerability",
                    "fixed_in": ["1.0.1"],
                    "aliases": [alias],
                    "link": "https://example.com/vuln-123",
                }
            ],
        },
    )

    status, renderables = await check_dependencies(
        APath(temp_uv_lock_file), Configuration(aliases=True)
    )

    assert status == 2
    for renderable in renderables:
        if not isinstance(renderable, Table):
            continue
        for column in renderable.columns:
            if column.header != "Aliases":
                continue
            cells = list(column.cells)
            assert len(cells) == 1
            cell = cells[0]
            assert isinstance(cell, Text)
            assert alias in cell.plain
            if expected_hyperlink is not None:
                assert expected_hyperlink in cell.markup



from datetime import timedelta
from pathlib import Path
from typing import Annotated, Optional

from pydantic import BaseModel, Field


DEFAULT_HTTPX_CACHE_TTL = 24.0 * 60.0 * 60.0  # Default cache time to 1 day


class CacheSettings(BaseModel):
    cache_path: Path = Path.home() / ".cache/uv-secure"
    disable_cache: bool = False
    ttl_seconds: Annotated[float, Field(ge=0.0, allow_inf_nan=False)] = (
        DEFAULT_HTTPX_CACHE_TTL
    )


class MaintainabilityCriteria(BaseModel):
    max_package_age: Optional[timedelta] = None
    forbid_yanked: bool = False


class Configuration(BaseModel):
    aliases: bool = False
    desc: bool = False
    ignore_vulnerabilities: Optional[set[str]] = None
    cache_settings: CacheSettings = CacheSettings()
    maintainability_criteria: MaintainabilityCriteria = MaintainabilityCriteria()


class OverrideConfiguration(BaseModel):
    aliases: Optional[bool] = None
    desc: Optional[bool] = None
    ignore_vulnerabilities: Optional[set[str]] = None
    forbid_yanked: Optional[bool] = None
    max_package_age: Optional[timedelta] = None
    disable_cache: Optional[bool] = None


def override_config(
    original_config: Configuration, overrides: OverrideConfiguration
) -> Configuration:
    """Override some configuration attributes from an override configuration

    Args:
        original_config: Original unmodified configuration
        overrides: Override attributes to override in original configuration

    Returns:
        Configuration with overridden attributes
    """

    new_configuration = original_config.model_copy()
    if overrides.aliases is not None:
        new_configuration.aliases = overrides.aliases
    if overrides.desc is not None:
        new_configuration.desc = overrides.desc
    if overrides.ignore_vulnerabilities is not None:
        new_configuration.ignore_vulnerabilities = overrides.ignore_vulnerabilities
    if overrides.forbid_yanked is not None:
        new_configuration.maintainability_criteria.forbid_yanked = (
            overrides.forbid_yanked
        )
    if overrides.max_package_age is not None:
        new_configuration.maintainability_criteria.max_package_age = (
            overrides.max_package_age
        )
    if overrides.disable_cache is not None:
        new_configuration.cache_settings.disable_cache = overrides.disable_cache

    return new_configuration



import sys

from anyio import Path
from pydantic import BaseModel
import typer


if sys.version_info >= (3, 11):
    import tomllib as toml
else:
    import tomli as toml


class Dependency(BaseModel):
    name: str
    version: str


async def parse_requirements_txt_file(file_path: Path) -> list[Dependency]:
    """Parse a requirements.txt file and extracts package PyPi dependencies"""
    data = await file_path.read_text()
    lines = data.splitlines()
    if (
        len(lines) == 0
        or lines[0].strip()
        != "# This file was autogenerated by uv via the following command:"
    ):
        typer.echo(
            f"Ignoring {file_path} as it doesn't appear to be a uv generated "
            "requirements.txt file"
        )
        return []
    dependencies = []
    for line in lines:
        if "==" not in line:
            continue
        name, version = line.split("==")
        dependencies.append(Dependency(name=name.strip(), version=version.strip()))
    return dependencies


async def parse_uv_lock_file(file_path: Path) -> list[Dependency]:
    """Parses a uv.lock TOML file and extracts package PyPi dependencies"""
    data = toml.loads(await file_path.read_text())

    package_data = data.get("package", [])
    return [
        Dependency(name=package["name"], version=package["version"])
        for package in package_data
        if package.get("source", {}).get("registry") == "https://pypi.org/simple"
    ]



from pathlib import Path
import sys
from typing import Optional

import typer

from uv_secure.__version__ import __version__
from uv_secure.dependency_checker import check_lock_files, RunStatus


app = typer.Typer()


def _version_callback(value: bool) -> None:
    if value:
        typer.echo(f"uv-secure {__version__}")
        raise typer.Exit()


_file_path_args = typer.Argument(
    None,
    help=(
        "Paths to the uv.lock or uv generated requirements.txt files or a single "
        "project root level directory (defaults to working directory if not set)"
    ),
)


_aliases_option = typer.Option(
    None,
    "--aliases",
    help="Flag whether to include vulnerability aliases in the vulnerabilities table",
)


_desc_option = typer.Option(
    None,
    "--desc",
    help=(
        "Flag whether to include vulnerability detailed description in the "
        "vulnerabilities table"
    ),
)

_disable_cache_option = typer.Option(
    None,
    "--disable-cache",
    help="Flag whether to disable caching for vulnerability http requests",
)

_forbid_yanked_option = typer.Option(
    None,
    "--forbid-yanked",
    help="Flag whether disallow yanked package versions from being dependencies",
)

_max_package_age_option = typer.Option(
    None, "--max-age-days", help="Maximum age threshold for packages in days"
)

_ignore_option = typer.Option(
    None,
    "--ignore",
    "-i",
    help="Comma-separated list of vulnerability IDs to ignore, e.g. VULN-123,VULN-456",
)

_config_option = typer.Option(
    None,
    "--config",
    help=(
        "Optional path to a configuration file (uv-secure.toml, .uv-secure.toml, or "
        "pyproject.toml)"
    ),
)

_version_option = typer.Option(
    None,
    "--version",
    callback=_version_callback,
    is_eager=True,
    help="Show the application's version",
)


@app.command()
def main(
    file_paths: Optional[list[Path]] = _file_path_args,
    aliases: Optional[bool] = _aliases_option,
    desc: Optional[bool] = _desc_option,
    disable_cache: Optional[bool] = _disable_cache_option,
    forbid_yanked: Optional[bool] = _forbid_yanked_option,
    max_package_age: Optional[int] = _max_package_age_option,
    ignore: Optional[str] = _ignore_option,
    config_path: Optional[Path] = _config_option,
    version: bool = _version_option,
) -> None:
    """Parse uv.lock files, check vulnerabilities, and display summary."""

    # Use uvloop or winloop if present
    try:
        if sys.platform in {"win32", "cygwin", "cli"}:
            from winloop import run
        else:
            from uvloop import run
    except ImportError:
        from asyncio import run

    run_status = run(
        check_lock_files(
            file_paths,
            aliases,
            desc,
            disable_cache,
            forbid_yanked,
            max_package_age,
            ignore,
            config_path,
        )
    )
    if run_status == RunStatus.MAINTENANCE_ISSUES_FOUND:
        raise typer.Exit(code=1)
    if run_status == RunStatus.VULNERABILITIES_FOUND:
        raise typer.Exit(code=2)
    if run_status == RunStatus.RUNTIME_ERROR:
        raise typer.Exit(code=3)


if __name__ == "__main__":
    app()  # pragma: no cover



from collections.abc import Iterable, Sequence
from typing import Union

from anyio import Path
from asyncer import create_task_group

from uv_secure.configuration import config_file_factory, Configuration


async def _search_file(directory: Path, filename: str) -> list[Path]:
    return [file_path async for file_path in directory.glob(f"**/{filename}")]


async def _find_files(
    directory: Path, filenames: Iterable[str]
) -> dict[str, list[Path]]:
    async with create_task_group() as tg:
        tasks = {
            filename: tg.soonify(_search_file)(directory, filename)
            for filename in filenames
        }

    return {filename: task.value for filename, task in tasks.items()}


async def _resolve_paths(file_paths: Sequence[Path]) -> list[Path]:
    async with create_task_group() as tg:
        tasks = [tg.soonify(path.resolve)() for path in file_paths]
    return [task.value for task in tasks]


def _get_root_dir(file_paths: Sequence[Path]) -> Path:
    if len(file_paths) == 1:
        return file_paths[0].parent

    split_paths = [list(rp.parts) for rp in file_paths]
    min_length = min(len(parts) for parts in split_paths)
    common_prefix_len = 0

    for part_idx in range(min_length):
        segment_set = {parts[part_idx] for parts in split_paths}
        if len(segment_set) == 1:
            common_prefix_len += 1
        else:
            break

    common_parts = split_paths[0][:common_prefix_len]
    return Path(*common_parts)


async def get_dependency_file_to_config_map(
    file_paths: Union[Path, list[Path]],
) -> dict[Path, Configuration]:
    """Get map of uv.lock files to their configurations.

    Using provided uv.lock files or root directory discover the uv.lock files and also
    find and map the nearest parent configuration for each uv.lock file.

    Args:
        file_paths: A list of uv.lock files or root directory

    Returns:
        A dictionary mapping uv.lock files to their nearest Configuration
    """
    if type(file_paths) is Path:
        root_dir = await file_paths.resolve()
        config_and_lock_files = await _find_files(
            root_dir,
            [
                "pyproject.toml",
                "uv-secure.toml",
                ".uv-secure.toml",
                "uv.lock",
                "requirements.txt",
            ],
        )
    else:
        resolved_paths = await _resolve_paths(file_paths)
        root_dir = _get_root_dir(resolved_paths)
        config_and_lock_files = await _find_files(
            root_dir, ["pyproject.toml", "uv-secure.toml", ".uv-secure.toml"]
        )
        config_and_lock_files["uv.lock"] = [
            path for path in resolved_paths if path.name == "uv.lock"
        ]
        config_and_lock_files["requirements.txt"] = [
            path for path in resolved_paths if path.name == "requirements.txt"
        ]

    config_file_paths = (
        config_and_lock_files["pyproject.toml"]
        + config_and_lock_files["uv-secure.toml"]
        + config_and_lock_files[".uv-secure.toml"]
    )

    async with create_task_group() as tg:
        config_futures = [
            tg.soonify(config_file_factory)(path) for path in config_file_paths
        ]
    configs = [future.value for future in config_futures]
    path_config_map = {
        p.parent: c for p, c in zip(config_file_paths, configs) if c is not None
    }

    dependency_file_paths = config_and_lock_files.get(
        "uv.lock", []
    ) + config_and_lock_files.get("requirements.txt", [])
    dependency_file_to_config_map: dict[Path, Configuration] = {}
    default_config = Configuration()
    for dependency_file in dependency_file_paths:
        current_dir = dependency_file.parent
        while True:
            found_config = path_config_map.get(current_dir)
            if found_config is not None or current_dir == root_dir:
                break
            current_dir = current_dir.parent

        if found_config is None:
            found_config = default_config
        dependency_file_to_config_map[dependency_file] = found_config
    return dependency_file_to_config_map



from collections.abc import Generator
from datetime import datetime
import os
from pathlib import Path
from textwrap import dedent
from typing import Callable
from zoneinfo import ZoneInfo

from freezegun import freeze_time
from httpx import Request, RequestError
import pytest
from pytest_httpx import HTTPXMock
from typer.testing import CliRunner

from uv_secure import app


runner = CliRunner()


@pytest.fixture
def temp_uv_requirements_txt_file(tmp_path: Path) -> Path:
    uv_requirements_txt_path = tmp_path / "requirements.txt"
    requirements_txt_data = """
        # This file was autogenerated by uv via the following command:
        #    uv pip compile pyproject.toml -o requirements.txt
        example-package==1.0.0
    """
    uv_requirements_txt_path.write_text(dedent(requirements_txt_data).strip())
    return uv_requirements_txt_path


@pytest.fixture
def temp_non_uv_requirements_txt_file(tmp_path: Path) -> Path:
    uv_requirements_txt_path = tmp_path / "requirements.txt"
    requirements_txt_data = """
        example-package==1.0.0
    """
    uv_requirements_txt_path.write_text(dedent(requirements_txt_data).strip())
    return uv_requirements_txt_path


@pytest.fixture
def temp_uv_secure_toml_file_ignored_vulnerability(tmp_path: Path) -> Path:
    uv_secure_toml_path = tmp_path / "uv-secure.toml"
    uv_lock_data = """
        ignore_vulnerabilities = ["VULN-123"]
    """
    uv_secure_toml_path.write_text(dedent(uv_lock_data).strip())
    return uv_secure_toml_path


@pytest.fixture
def temp_uv_secure_toml_file_all_columns_enabled(tmp_path: Path) -> Path:
    uv_secure_toml_path = tmp_path / "uv-secure.toml"
    uv_lock_data = """
        aliases = true
        desc = true
    """
    uv_secure_toml_path.write_text(dedent(uv_lock_data).strip())
    return uv_secure_toml_path


@pytest.fixture
def temp_uv_secure_toml_file_all_columns_and_maintenance_issues_enabled(
    tmp_path: Path,
) -> Path:
    uv_secure_toml_path = tmp_path / "uv-secure.toml"
    uv_lock_data = """
        aliases = true
        desc = true

        [maintainability_criteria]
        max_package_age = "P1000D"
        forbid_yanked = true
    """
    uv_secure_toml_path.write_text(dedent(uv_lock_data).strip())
    return uv_secure_toml_path


@pytest.fixture
def temp_uv_secure_toml_file_custom_caching(tmp_path: Path) -> Path:
    uv_secure_toml_path = tmp_path / "uv-secure.toml"
    uv_lock_data = f"""
        aliases = true
        desc = true

        [cache_settings]
        cache_path = "{(tmp_path / ".uv-secure").as_posix()}"
        ttl_seconds = 60.0
    """
    uv_secure_toml_path.write_text(dedent(uv_lock_data).strip())
    return uv_secure_toml_path


@pytest.fixture
def temp_dot_uv_secure_toml_file(tmp_path: Path) -> Path:
    uv_secure_toml_path = tmp_path / ".uv-secure.toml"
    uv_lock_data = ""
    uv_secure_toml_path.write_text(uv_lock_data)
    return uv_secure_toml_path


@pytest.fixture
def temp_nested_uv_secure_toml_file_ignored_vulnerability(tmp_path: Path) -> Path:
    nested_uv_lock_path = tmp_path / "nested_project"
    uv_secure_toml_path = nested_uv_lock_path / "uv-secure.toml"
    uv_lock_data = """
    ignore_vulnerabilities = ["VULN-123"]
    """
    uv_secure_toml_path.write_text(dedent(uv_lock_data).strip())
    return uv_secure_toml_path


@pytest.fixture
def temp_pyproject_toml_file(tmp_path: Path) -> Path:
    pyproject_toml_path = tmp_path / "pyproject.toml"
    uv_lock_data = """
    [tool.uv-secure]
    """
    pyproject_toml_path.write_text(dedent(uv_lock_data).strip())
    return pyproject_toml_path


@pytest.fixture
def temp_pyproject_toml_file_extra_columns_enabled(tmp_path: Path) -> Path:
    pyproject_toml_path = tmp_path / "pyproject.toml"
    uv_lock_data = """
    [tool.uv-secure]
    aliases = true
    desc = true
    """
    pyproject_toml_path.write_text(dedent(uv_lock_data).strip())
    return pyproject_toml_path


@pytest.fixture
def temp_pyproject_toml_file_ignored_vulnerability(tmp_path: Path) -> Path:
    pyproject_toml_path = tmp_path / "pyproject.toml"
    uv_lock_data = """
    [tool.uv-secure]
    ignore_vulnerabilities = ["VULN-123"]
    """
    pyproject_toml_path.write_text(dedent(uv_lock_data).strip())
    return pyproject_toml_path


@pytest.fixture
def temp_nested_pyproject_toml_file_no_config(tmp_path: Path) -> Path:
    pyproject_toml_path = tmp_path / "nested_project" / "pyproject.toml"
    uv_lock_data = ""
    pyproject_toml_path.write_text(uv_lock_data)
    return pyproject_toml_path


@pytest.fixture
def temp_nested_uv_lock_file(tmp_path: Path) -> Path:
    nested_uv_lock_path = tmp_path / "nested_project"
    nested_uv_lock_path.mkdir()
    uv_lock_path = nested_uv_lock_path / "uv.lock"
    uv_lock_data = """
    [[package]]
    name = "example-package"
    version = "2.0.0"
    source = { registry = "https://pypi.org/simple" }
    """
    uv_lock_path.write_text(dedent(uv_lock_data).strip())
    return uv_lock_path


@pytest.fixture
def temp_double_nested_uv_lock_file(tmp_path: Path) -> Path:
    double_nested_uv_lock_path = tmp_path / "nested_project" / "double_nested_project"
    double_nested_uv_lock_path.mkdir(parents=True)
    uv_lock_path = double_nested_uv_lock_path / "uv.lock"
    uv_lock_data = """
    [[package]]
    name = "example-package"
    version = "2.0.0"
    source = { registry = "https://pypi.org/simple" }
    """
    uv_lock_path.write_text(dedent(uv_lock_data).strip())
    return uv_lock_path


@pytest.fixture
def no_vulnerabilities_response(httpx_mock: HTTPXMock) -> HTTPXMock:
    httpx_mock.add_response(
        url="https://pypi.org/pypi/example-package/1.0.0/json",
        json={
            "info": {
                "author_email": "example@example.com",
                "classifiers": [],
                "description": "A minimal package",
                "description_content_type": "text/plain",
                "downloads": {"last_day": None, "last_month": None, "last_week": None},
                "name": "example-package",
                "project_urls": {},
                "provides_extra": [],
                "release_url": "https://pypi.org/project/example-package/1.0.0/",
                "requires_python": ">=3.9",
                "summary": "A minimal package example",
                "version": "1.0.0",
                "yanked": False,
            },
            "last_serial": 1,
            "urls": [],
            "vulnerabilities": [],
        },
    )
    return httpx_mock


@pytest.fixture
def one_vulnerability_response(httpx_mock: HTTPXMock) -> HTTPXMock:
    httpx_mock.add_response(
        url="https://pypi.org/pypi/example-package/1.0.0/json",
        json={
            "info": {
                "author_email": "example@example.com",
                "classifiers": [],
                "description": "A minimal package",
                "description_content_type": "text/plain",
                "downloads": {"last_day": None, "last_month": None, "last_week": None},
                "name": "example-package",
                "project_urls": {},
                "provides_extra": [],
                "release_url": "https://pypi.org/project/example-package/1.0.0/",
                "requires_python": ">=3.9",
                "summary": "A minimal package example",
                "version": "1.0.0",
                "yanked": False,
            },
            "last_serial": 1,
            "urls": [],
            "vulnerabilities": [
                {
                    "aliases": ["CVE-2024-12345"],
                    "id": "VULN-123",
                    "details": "A critical vulnerability in example-package.",
                    "fixed_in": ["1.0.1"],
                    "link": "https://example.com/vuln-123",
                }
            ],
        },
    )
    return httpx_mock


@pytest.fixture
def old_yanked_package_response(httpx_mock: HTTPXMock) -> HTTPXMock:
    httpx_mock.add_response(
        url="https://pypi.org/pypi/example-package/1.0.0/json",
        json={
            "info": {
                "author_email": "example@example.com",
                "classifiers": [],
                "description": "A minimal package",
                "description_content_type": "text/plain",
                "downloads": {"last_day": None, "last_month": None, "last_week": None},
                "name": "example-package",
                "project_urls": {},
                "provides_extra": [],
                "release_url": "https://pypi.org/project/example-package/1.0.0/",
                "requires_python": ">=3.9",
                "summary": "A minimal package example",
                "version": "1.0.0",
                "yanked": True,
                "yanked_reason": "Broken API",
            },
            "last_serial": 1,
            "urls": [
                {
                    "comment_text": "",
                    "digests": {
                        "blake2b_256": (
                            "0bf785273299ab57117850cc0a936c64151171fac4da49bc6fba0dad98"
                            "4a7c5f"
                        ),
                        "md5": "8626f021f29631950dfad7b4c6435fc4",
                        "sha256": (
                            "8a3df80e2b2378aef598a83c1392efd47967afec4242021a0b06b4c7cb"
                            "c61a92"
                        ),
                    },
                    "downloads": -1,
                    "filename": "example-package-1.0.0-py3-none-any.whl",
                    "has_sig": False,
                    "md5_digest": "8626f021f29631950dfad7b4c6435fc4",
                    "packagetype": "bdist_wheel",
                    "python_version": "py3",
                    "requires_python": ">=3.7",
                    "size": 15662,
                    "upload_time": "2021-01-19T23:44:28",
                    "upload_time_iso_8601": "2021-01-19T23:44:28.833863Z",
                    "url": (
                        "https://files.pythonhosted.org/packages/0b/f7/"
                        "85273299ab57117850cc0a936c64151171fac4da49bc6fba0dad984a7c5f/"
                        "example-package-1.0.0-py3-none-any.whl"
                    ),
                    "yanked": True,
                    "yanked_reason": "Broken API",
                }
            ],
            "vulnerabilities": [],
        },
    )
    return httpx_mock


@pytest.fixture
def yanked_package_no_reason_given_response(httpx_mock: HTTPXMock) -> HTTPXMock:
    httpx_mock.add_response(
        url="https://pypi.org/pypi/example-package/1.0.0/json",
        json={
            "info": {
                "author_email": "example@example.com",
                "classifiers": [],
                "description": "A minimal package",
                "description_content_type": "text/plain",
                "downloads": {"last_day": None, "last_month": None, "last_week": None},
                "name": "example-package",
                "project_urls": {},
                "provides_extra": [],
                "release_url": "https://pypi.org/project/example-package/1.0.0/",
                "requires_python": ">=3.9",
                "summary": "A minimal package example",
                "version": "1.0.0",
                "yanked": True,
            },
            "last_serial": 1,
            "urls": [
                {
                    "comment_text": "",
                    "digests": {
                        "blake2b_256": (
                            "0bf785273299ab57117850cc0a936c64151171fac4da49bc6fba0dad98"
                            "4a7c5f"
                        ),
                        "md5": "8626f021f29631950dfad7b4c6435fc4",
                        "sha256": (
                            "8a3df80e2b2378aef598a83c1392efd47967afec4242021a0b06b4c7cb"
                            "c61a92"
                        ),
                    },
                    "downloads": -1,
                    "filename": "example-package-1.0.0-py3-none-any.whl",
                    "has_sig": False,
                    "md5_digest": "8626f021f29631950dfad7b4c6435fc4",
                    "packagetype": "bdist_wheel",
                    "python_version": "py3",
                    "requires_python": ">=3.7",
                    "size": 15662,
                    "upload_time": "2024-01-19T23:44:28",
                    "upload_time_iso_8601": "2024-01-19T23:44:28.833863Z",
                    "url": (
                        "https://files.pythonhosted.org/packages/0b/f7/"
                        "85273299ab57117850cc0a936c64151171fac4da49bc6fba0dad984a7c5f/"
                        "example-package-1.0.0-py3-none-any.whl"
                    ),
                    "yanked": True,
                }
            ],
            "vulnerabilities": [],
        },
    )
    return httpx_mock


@pytest.fixture
def old_yanked_package_with_vulnerability_response(httpx_mock: HTTPXMock) -> HTTPXMock:
    httpx_mock.add_response(
        url="https://pypi.org/pypi/example-package/1.0.0/json",
        json={
            "info": {
                "author_email": "example@example.com",
                "classifiers": [],
                "description": "A minimal package",
                "description_content_type": "text/plain",
                "downloads": {"last_day": None, "last_month": None, "last_week": None},
                "name": "example-package",
                "project_urls": {},
                "provides_extra": [],
                "release_url": "https://pypi.org/project/example-package/1.0.0/",
                "requires_python": ">=3.9",
                "summary": "A minimal package example",
                "version": "1.0.0",
                "yanked": True,
                "yanked_reason": "Broken API",
            },
            "last_serial": 1,
            "urls": [
                {
                    "comment_text": "",
                    "digests": {
                        "blake2b_256": (
                            "0bf785273299ab57117850cc0a936c64151171fac4da49bc6fba0dad98"
                            "4a7c5f"
                        ),
                        "md5": "8626f021f29631950dfad7b4c6435fc4",
                        "sha256": (
                            "8a3df80e2b2378aef598a83c1392efd47967afec4242021a0b06b4c7cb"
                            "c61a92"
                        ),
                    },
                    "downloads": -1,
                    "filename": "example-package-1.0.0-py3-none-any.whl",
                    "has_sig": False,
                    "md5_digest": "8626f021f29631950dfad7b4c6435fc4",
                    "packagetype": "bdist_wheel",
                    "python_version": "py3",
                    "requires_python": ">=3.7",
                    "size": 15662,
                    "upload_time": "2021-01-19T23:44:28",
                    "upload_time_iso_8601": "2021-01-19T23:44:28.833863Z",
                    "url": (
                        "https://files.pythonhosted.org/packages/0b/f7/"
                        "85273299ab57117850cc0a936c64151171fac4da49bc6fba0dad984a7c5f/"
                        "example-package-1.0.0-py3-none-any.whl"
                    ),
                    "yanked": True,
                    "yanked_reason": "Broken API",
                }
            ],
            "vulnerabilities": [
                {
                    "aliases": ["CVE-2024-12345"],
                    "id": "VULN-123",
                    "details": "A critical vulnerability in example-package.",
                    "fixed_in": ["1.0.1"],
                    "link": "https://example.com/vuln-123",
                }
            ],
        },
    )
    return httpx_mock


@pytest.fixture
def temp_uv_lock_file_jinja2(tmp_path: Path) -> Path:
    """Fixture to create a temporary uv.lock file with a single jinja2 dependency."""
    uv_lock_path = tmp_path / "uv.lock"
    uv_lock_data = """
    [[package]]
    name = "jinja2"
    version = "3.1.4"
    source = { registry = "https://pypi.org/simple" }
    """
    uv_lock_path.write_text(dedent(uv_lock_data).strip())
    return uv_lock_path


@pytest.fixture
def jinja2_two_longer_vulnerability_responses(httpx_mock: HTTPXMock) -> HTTPXMock:
    httpx_mock.add_response(
        url="https://pypi.org/pypi/jinja2/3.1.4/json",
        json={
            "info": {
                "author_email": "example@example.com",
                "classifiers": [],
                "description": "Jinja2 templating",
                "description_content_type": "text/plain",
                "downloads": {"last_day": None, "last_month": None, "last_week": None},
                "name": "jinja2",
                "project_urls": {},
                "provides_extra": [],
                "release_url": "https://pypi.org/project/jinja2/3.1.4/",
                "requires_python": ">=3.9",
                "summary": "Jinja2 templating",
                "version": "3.1.4",
                "yanked": False,
            },
            "last_serial": 1,
            "urls": [],
            "vulnerabilities": [
                {
                    "aliases": ["CVE-2024-56326"],
                    "details": (
                        "An oversight in how the Jinja sandboxed environment detects "
                        "calls to `str.format` allows an attacker that controls the "
                        "content of a template to execute arbitrary Python code.\n\nTo "
                        "exploit the vulnerability, an attacker needs to control the "
                        "content of a template. Whether that is the case depends on "
                        "the type of application using Jinja. This vulnerability "
                        "impacts users of applications which execute untrusted "
                        "templates.\n\nJinja's sandbox does catch calls to "
                        "`str.format` and ensures they don't escape the sandbox. "
                        "However, it's possible to store a reference to a malicious "
                        "string's `format` method, then pass that to a filter that "
                        "calls it. No such filters are built-in to Jinja, but could be "
                        "present through custom filters in an application. After the "
                        "fix, such indirect calls are also handled by the sandbox."
                    ),
                    "fixed_in": ["3.1.5"],
                    "id": "GHSA-q2x7-8rv6-6q7h",
                    "link": "https://osv.dev/vulnerability/GHSA-q2x7-8rv6-6q7h",
                    "source": "osv",
                    "summary": None,
                    "withdrawn": None,
                },
                {
                    "aliases": ["CVE-2024-56201"],
                    "details": (
                        "A bug in the Jinja compiler allows an attacker that controls "
                        "both the content and filename of a template to execute "
                        "arbitrary Python code, regardless of if Jinja's sandbox is "
                        "used.\n\nTo exploit the vulnerability, an attacker needs to "
                        "control both the filename and the contents of a template. "
                        "Whether that is the case depends on the type of application "
                        "using Jinja. This vulnerability impacts users of applications "
                        "which execute untrusted templates where the template author "
                        "can also choose the template filename."
                    ),
                    "fixed_in": ["3.1.5"],
                    "id": "GHSA-gmj6-6f8f-6699",
                    "link": "https://osv.dev/vulnerability/GHSA-gmj6-6f8f-6699",
                    "source": "osv",
                    "summary": None,
                    "withdrawn": None,
                },
            ],
        },
    )
    return httpx_mock


@pytest.fixture
def one_vulnerability_response_v2(httpx_mock: HTTPXMock) -> HTTPXMock:
    httpx_mock.add_response(
        url="https://pypi.org/pypi/example-package/2.0.0/json",
        json={
            "info": {
                "author_email": "example@example.com",
                "classifiers": [],
                "description": "A minimal package",
                "description_content_type": "text/plain",
                "downloads": {"last_day": None, "last_month": None, "last_week": None},
                "name": "example-package",
                "project_urls": {},
                "provides_extra": [],
                "release_url": "https://pypi.org/project/example-package/2.0.0/",
                "requires_python": ">=3.9",
                "summary": "A minimal package example",
                "version": "2.0.0",
                "yanked": False,
            },
            "last_serial": 1,
            "urls": [],
            "vulnerabilities": [
                {
                    "id": "VULN-123",
                    "details": "A critical vulnerability in example-package.",
                    "fixed_in": ["2.0.1"],
                    "link": "https://example.com/vuln-123",
                }
            ],
        },
    )
    return httpx_mock


@pytest.fixture
def package_version_not_found_response(httpx_mock: HTTPXMock) -> HTTPXMock:
    httpx_mock.add_response(
        url="https://pypi.org/pypi/example-package/1.0.0/json", status_code=404
    )
    return httpx_mock


@pytest.fixture
def missing_vulnerability_response(httpx_mock: HTTPXMock) -> HTTPXMock:
    httpx_mock.add_exception(
        RequestError(
            "Request failed",
            request=Request("GET", "https://pypi.org/pypi/example-package/1.0.0/json"),
        )
    )
    return httpx_mock


@pytest.fixture(scope="session", autouse=True)
def wide_console() -> Generator:
    mp = pytest.MonkeyPatch()
    mp.setenv("COLUMNS", "200")
    yield mp
    mp.undo()


@pytest.fixture
def set_console_width(monkeypatch: pytest.MonkeyPatch) -> Callable[[int], None]:
    def _set_width(width: int) -> None:
        monkeypatch.setenv("COLUMNS", str(width))

    return _set_width


def test_app_version() -> None:
    result = runner.invoke(app, "--version")
    assert result.exit_code == 0
    assert "uv-secure " in result.output


def test_bad_file_name() -> None:
    result = runner.invoke(app, "i_dont_exist.txt")
    assert result.exit_code == 3
    assert "Error" in result.output


def test_missing_file(tmp_path: Path) -> None:
    result = runner.invoke(app, [str(tmp_path / "uv.lock")])
    assert result.exit_code == 3
    assert "Error" in result.output


def test_non_uv_requirements_txt_file(temp_non_uv_requirements_txt_file: Path) -> None:
    result = runner.invoke(app, [str(temp_non_uv_requirements_txt_file)])

    assert result.exit_code == 0
    assert "doesn't appear to be a uv generated requirements.txt file" in result.output


def test_app_no_vulnerabilities(
    temp_uv_lock_file: Path, no_vulnerabilities_response: HTTPXMock
) -> None:
    result = runner.invoke(app, [str(temp_uv_lock_file), "--disable-cache"])

    assert result.exit_code == 0
    assert "No vulnerabilities detected!" in result.output
    assert "Checked: 1 dependency" in result.output
    assert "All dependencies appear safe!" in result.output


def test_app_no_vulnerabilities_requirements_txt(
    temp_uv_requirements_txt_file: Path, no_vulnerabilities_response: HTTPXMock
) -> None:
    result = runner.invoke(app, [str(temp_uv_requirements_txt_file), "--disable-cache"])

    assert result.exit_code == 0
    assert "No vulnerabilities detected!" in result.output
    assert "Checked: 1 dependency" in result.output
    assert "All dependencies appear safe!" in result.output


def test_app_no_vulnerabilities_requirements_txt_no_specified_path(
    tmp_path: Path,
    temp_uv_requirements_txt_file: Path,
    no_vulnerabilities_response: HTTPXMock,
) -> None:
    os.chdir(tmp_path)
    result = runner.invoke(app, "--disable-cache")

    assert result.exit_code == 0
    assert "No vulnerabilities detected!" in result.output
    assert "Checked: 1 dependency" in result.output
    assert "All dependencies appear safe!" in result.output


def test_app_no_vulnerabilities_relative_lock_file_path(
    tmp_path: Path, temp_uv_lock_file: Path, no_vulnerabilities_response: HTTPXMock
) -> None:
    os.chdir(tmp_path)
    result = runner.invoke(app, ["uv.lock", "--disable-cache"])

    assert result.exit_code == 0
    assert "No vulnerabilities detected!" in result.output
    assert "Checked: 1 dependency" in result.output
    assert "All dependencies appear safe!" in result.output


def test_app_no_vulnerabilities_relative_no_specified_path(
    tmp_path: Path, temp_uv_lock_file: Path, no_vulnerabilities_response: HTTPXMock
) -> None:
    os.chdir(tmp_path)
    result = runner.invoke(app, "--disable-cache")

    assert result.exit_code == 0
    assert "No vulnerabilities detected!" in result.output
    assert "Checked: 1 dependency" in result.output
    assert "All dependencies appear safe!" in result.output


@freeze_time(datetime(2025, 1, 30, tzinfo=ZoneInfo("UTC")))
def test_app_maintenance_issues_cli_args(
    temp_uv_lock_file: Path, old_yanked_package_response: HTTPXMock
) -> None:
    result = runner.invoke(
        app,
        [
            str(temp_uv_lock_file),
            "--forbid-yanked",
            "--max-age-days",
            "1000",
            "--disable-cache",
        ],
    )

    assert result.exit_code == 1
    assert "Maintenance Issues detected!" in result.output
    assert "Checked: 1 dependency" in result.output
    assert "Issues: 1 issue" in result.output
    assert "Maintenance Issues" in result.output
    assert "Broken API" in result.output
    assert "4 years and 11.01 days" in result.output


@freeze_time(datetime(2025, 1, 30, tzinfo=ZoneInfo("UTC")))
def test_app_yanked_no_reason_cli_args(
    temp_uv_lock_file: Path, yanked_package_no_reason_given_response: HTTPXMock
) -> None:
    result = runner.invoke(
        app, [str(temp_uv_lock_file), "--forbid-yanked", "--disable-cache"]
    )

    assert result.exit_code == 1
    assert "Maintenance Issues detected!" in result.output
    assert "Checked: 1 dependency" in result.output
    assert "Issues: 1 issue" in result.output
    assert "Maintenance Issues" in result.output
    assert "Unknown" in result.output
    assert "1 year and 11.01 days" in result.output


def test_app_failed_vulnerability_request(
    temp_uv_lock_file: Path, missing_vulnerability_response: HTTPXMock
) -> None:
    result = runner.invoke(app, [str(temp_uv_lock_file), "--disable-cache"])

    assert result.exit_code == 0
    assert (
        "Error: name='example-package' version='1.0.0' raised exception: Request failed"
    ) in result.output
    assert "No vulnerabilities detected!" in result.output
    assert "Checked: 1 dependency" in result.output
    assert "All dependencies appear safe!" in result.output


def test_app_package_not_found(
    temp_uv_lock_file: Path, package_version_not_found_response: HTTPXMock
) -> None:
    result = runner.invoke(app, [str(temp_uv_lock_file), "--disable-cache"])

    assert result.exit_code == 0
    assert (
        "Error: name='example-package' version='1.0.0' raised exception: Client "
        "error '404 Not Found' for url "
        "'https://pypi.org/pypi/example-package/1.0.0/json'"
    ) in result.output
    assert "No vulnerabilities detected!" in result.output
    assert "Checked: 1 dependency" in result.output
    assert "All dependencies appear safe!" in result.output


@pytest.mark.parametrize(
    "extra_cli_args",
    [
        pytest.param([], id="Default arguments"),
        pytest.param(["--aliases"], id="Add Aliases column"),
        pytest.param(["--desc"], id="Add details column"),
        pytest.param(["--aliases", "--desc"], id="Add details column"),
        pytest.param(
            ["--forbid-yanked", "--max-age-days", "1000"], id="Maintenance criteria"
        ),
    ],
)
@freeze_time(datetime(2025, 1, 30, tzinfo=ZoneInfo("UTC")))
def test_check_dependencies_with_vulnerability(
    extra_cli_args: list[str],
    temp_uv_lock_file: Path,
    one_vulnerability_response: HTTPXMock,
) -> None:
    """Test check_dependencies with a single dependency and a single vulnerability."""
    result = runner.invoke(
        app, [str(temp_uv_lock_file), *extra_cli_args, "--disable-cache"]
    )

    assert result.exit_code == 2
    assert "Vulnerabilities detected!" in result.output
    assert "Checked: 1 dependency" in result.output
    assert "Vulnerable: 1 vulnerability" in result.output
    assert "example-package" in result.output
    assert "1.0.0" in result.output
    assert "VULN-123" in result.output
    assert "1.0.1" in result.output
    if "--aliases" in extra_cli_args:
        assert "Aliases" in result.output
        assert "CVE-2024-12345" in result.output
    if "--desc" in extra_cli_args:
        assert "Details" in result.output
        assert "A critical vulnerability in example-package.  " in result.output


def test_check_dependencies_with_vulnerability_narrow_console_vulnerability_ids_visible(
    temp_uv_lock_file_jinja2: Path,
    jinja2_two_longer_vulnerability_responses: HTTPXMock,
    set_console_width: Callable[[int], None],
) -> None:
    """Test check_dependencies with a single dependency and a single vulnerability."""
    set_console_width(80)
    result = runner.invoke(
        app, [str(temp_uv_lock_file_jinja2), "--aliases", "--desc", "--disable-cache"]
    )

    assert result.exit_code == 2
    assert "GHSA-q2x7-8rv6-6q7h" in result.output
    assert "GHSA-gmj6-6f8f-6699" in result.output


def test_check_dependencies_with_two_longer_vulnerabilities(
    temp_uv_lock_file_jinja2: Path, jinja2_two_longer_vulnerability_responses: HTTPXMock
) -> None:
    """Test check_dependencies with a single dependency and a single vulnerability."""
    result = runner.invoke(app, [str(temp_uv_lock_file_jinja2), "--disable-cache"])

    assert result.exit_code == 2
    assert "Vulnerabilities detected!" in result.output
    assert "Checked: 1 dependency" in result.output
    assert "Vulnerable: 2 vulnerabilities" in result.output
    assert result.output.count("jinja2") == 2
    assert result.output.count("3.1.4") == 2
    assert result.output.count("3.1.5") == 2
    assert "GHSA-q2x7-8rv6-6q7h" in result.output
    assert "GHSA-gmj6-6f8f-6699" in result.output


def test_app_with_arg_ignored_vulnerability(
    temp_uv_lock_file: Path, one_vulnerability_response: HTTPXMock
) -> None:
    result = runner.invoke(
        app, [str(temp_uv_lock_file), "--ignore", "VULN-123", "--disable-cache"]
    )

    assert result.exit_code == 0
    assert "No vulnerabilities detected!" in result.output
    assert "Checked: 1 dependency" in result.output
    assert "All dependencies appear safe!" in result.output


def test_check_dependencies_with_vulnerability_pyproject_all_columns_configured(
    temp_uv_lock_file: Path,
    temp_pyproject_toml_file_extra_columns_enabled: Path,
    one_vulnerability_response: HTTPXMock,
) -> None:
    """Test check_dependencies with a single dependency and a single vulnerability."""
    result = runner.invoke(app, [str(temp_uv_lock_file), "--disable-cache"])

    assert result.exit_code == 2
    assert "Vulnerabilities detected!" in result.output
    assert "Checked: 1 dependency" in result.output
    assert "Vulnerable: 1 vulnerability" in result.output
    assert "example-package" in result.output
    assert "1.0.0" in result.output
    assert "VULN-123" in result.output
    assert "1.0.1" in result.output
    assert "Aliases" in result.output
    assert "CVE-2024-12345" in result.output
    assert "Details" in result.output
    assert "A critical vulnerability in example-package.  " in result.output


def test_check_dependencies_with_vulnerability_uv_secure_all_columns_configured(
    temp_uv_lock_file: Path,
    temp_uv_secure_toml_file_all_columns_enabled: Path,
    one_vulnerability_response: HTTPXMock,
) -> None:
    result = runner.invoke(app, [str(temp_uv_lock_file), "--disable-cache"])

    assert result.exit_code == 2
    assert "Vulnerabilities detected!" in result.output
    assert "Checked: 1 dependency" in result.output
    assert "Vulnerable: 1 vulnerability" in result.output
    assert "example-package" in result.output
    assert "1.0.0" in result.output
    assert "VULN-123" in result.output
    assert "1.0.1" in result.output
    assert "Aliases" in result.output
    assert "CVE-2024-12345" in result.output
    assert "Details" in result.output
    assert "A critical vulnerability in example-package.  " in result.output


@freeze_time(datetime(2025, 1, 30, tzinfo=ZoneInfo("UTC")))
def test_check_dependencies_with_vulnerability_and_maintenance_issues_uv_secure(
    temp_uv_lock_file: Path,
    temp_uv_secure_toml_file_all_columns_and_maintenance_issues_enabled: Path,
    old_yanked_package_with_vulnerability_response: HTTPXMock,
) -> None:
    result = runner.invoke(app, [str(temp_uv_lock_file), "--disable-cache"])

    assert result.exit_code == 2
    assert "Vulnerabilities detected!" in result.output
    assert "Checked: 1 dependency" in result.output
    assert "Vulnerable: 1 vulnerability" in result.output
    assert "example-package" in result.output
    assert "1.0.0" in result.output
    assert "VULN-123" in result.output
    assert "1.0.1" in result.output
    assert "Aliases" in result.output
    assert "CVE-2024-12345" in result.output
    assert "Details" in result.output
    assert "A critical vulnerability in example-package." in result.output
    assert "Maintenance Issues detected!" in result.output
    assert "Checked: 1 dependency" in result.output
    assert "Issues: 1 issue" in result.output
    assert "Maintenance Issues" in result.output
    assert "Broken API" in result.output
    assert "4 years and 11.01 days" in result.output


def test_check_dependencies_with_custom_caching(
    temp_uv_lock_file: Path,
    temp_uv_secure_toml_file_custom_caching: Path,
    no_vulnerabilities_response: HTTPXMock,
) -> None:
    result = runner.invoke(app, [str(temp_uv_lock_file), "--disable-cache"])
    assert result.exit_code == 0
    assert "No vulnerabilities detected!" in result.output
    assert "Checked: 1 dependency" in result.output
    assert "All dependencies appear safe!" in result.output


def test_check_dependencies_with_vulnerability_pyproject_toml_cli_argument_override(
    temp_uv_lock_file: Path,
    temp_pyproject_toml_file_ignored_vulnerability: Path,
    one_vulnerability_response: HTTPXMock,
) -> None:
    result = runner.invoke(
        app,
        [str(temp_uv_lock_file), "--ignore", "VULN-NOT-HERE", "--aliases", "--desc"],
        "--disable-cache",
    )

    assert "Vulnerabilities detected!" in result.output
    assert "Checked: 1 dependency" in result.output
    assert "Vulnerable: 1 vulnerability" in result.output
    assert "example-package" in result.output
    assert "1.0.0" in result.output
    assert "VULN-123" in result.output
    assert "1.0.1" in result.output
    assert "Aliases" in result.output
    assert "CVE-2024-12345" in result.output
    assert "Details" in result.output
    assert "A critical vulnerability in example-package.  " in result.output


def test_app_with_uv_secure_toml_ignored_vulnerability(
    temp_uv_lock_file: Path,
    temp_uv_secure_toml_file_ignored_vulnerability: Path,
    one_vulnerability_response: HTTPXMock,
) -> None:
    result = runner.invoke(
        app,
        [
            str(temp_uv_lock_file),
            "--config",
            temp_uv_secure_toml_file_ignored_vulnerability,
        ],
        "--disable-cache",
    )

    assert result.exit_code == 0
    assert "No vulnerabilities detected!" in result.output
    assert "Checked: 1 dependency" in result.output
    assert "All dependencies appear safe!" in result.output


def test_app_with_pyproject_toml_ignored_vulnerability(
    temp_uv_lock_file: Path,
    temp_pyproject_toml_file_ignored_vulnerability: Path,
    one_vulnerability_response: HTTPXMock,
) -> None:
    result = runner.invoke(
        app,
        [
            str(temp_uv_lock_file),
            "--config",
            temp_pyproject_toml_file_ignored_vulnerability,
        ],
        "--disable-cache",
    )

    assert result.exit_code == 0
    assert "No vulnerabilities detected!" in result.output
    assert "Checked: 1 dependency" in result.output
    assert "All dependencies appear safe!" in result.output


def test_app_multiple_lock_files_no_vulnerabilities(
    temp_uv_lock_file: Path, temp_nested_uv_lock_file: Path, httpx_mock: HTTPXMock
) -> None:
    httpx_mock.add_response(
        url="https://pypi.org/pypi/example-package/1.0.0/json",
        json={"vulnerabilities": []},
    )
    httpx_mock.add_response(
        url="https://pypi.org/pypi/example-package/2.0.0/json",
        json={"vulnerabilities": []},
    )

    result = runner.invoke(
        app, [str(temp_uv_lock_file), str(temp_nested_uv_lock_file), "--disable-cache"]
    )

    assert result.exit_code == 0
    assert result.output.count("No vulnerabilities detected!") == 2
    assert result.output.count("Checked: 1 dependency") == 2
    assert result.output.count("All dependencies appear safe!") == 2
    assert result.output.count("nested_project") == 1


def test_app_multiple_lock_files_one_vulnerabilities(
    temp_uv_lock_file: Path,
    temp_nested_uv_lock_file: Path,
    no_vulnerabilities_response: HTTPXMock,
    one_vulnerability_response_v2: HTTPXMock,
) -> None:
    result = runner.invoke(
        app, [str(temp_uv_lock_file), str(temp_nested_uv_lock_file), "--disable-cache"]
    )
    assert result.exit_code == 2
    assert result.output.count("No vulnerabilities detected!") == 1
    assert result.output.count("Vulnerabilities detected!") == 1


def test_app_multiple_lock_files_one_nested_ignored_vulnerability(
    tmp_path: Path,
    temp_uv_lock_file: Path,
    temp_nested_uv_lock_file: Path,
    temp_dot_uv_secure_toml_file: Path,
    temp_nested_uv_secure_toml_file_ignored_vulnerability: Path,
    no_vulnerabilities_response: HTTPXMock,
    one_vulnerability_response_v2: HTTPXMock,
) -> None:
    result = runner.invoke(app, [str(tmp_path), "--disable-cache"])

    assert result.exit_code == 0
    assert result.output.count("No vulnerabilities detected!") == 2
    assert result.output.count("Checked: 1 dependency") == 2
    assert result.output.count("All dependencies appear safe!") == 2
    assert result.output.count("nested_project") == 1


def test_app_multiple_lock_files_no_root_config_one_nested_ignored_vulnerability(
    tmp_path: Path,
    temp_uv_lock_file: Path,
    temp_double_nested_uv_lock_file: Path,
    temp_nested_uv_secure_toml_file_ignored_vulnerability: Path,
    no_vulnerabilities_response: HTTPXMock,
    one_vulnerability_response_v2: HTTPXMock,
) -> None:
    result = runner.invoke(app, [str(tmp_path), "--disable-cache"])

    assert result.exit_code == 0
    assert result.output.count("No vulnerabilities detected!") == 2
    assert result.output.count("Checked: 1 dependency") == 2
    assert result.output.count("All dependencies appear safe!") == 2
    assert result.output.count("nested_project") == 2


def test_app_multiple_lock_files_one_nested_ignored_vulnerability_pass_lock_files(
    tmp_path: Path,
    temp_uv_lock_file: Path,
    temp_double_nested_uv_lock_file: Path,
    temp_nested_uv_secure_toml_file_ignored_vulnerability: Path,
    no_vulnerabilities_response: HTTPXMock,
    one_vulnerability_response_v2: HTTPXMock,
) -> None:
    result = runner.invoke(
        app,
        [str(temp_uv_lock_file), str(temp_double_nested_uv_lock_file)],
        "--disable-cache",
    )

    assert result.exit_code == 0
    assert result.output.count("No vulnerabilities detected!") == 2
    assert result.output.count("Checked: 1 dependency") == 2
    assert result.output.count("All dependencies appear safe!") == 2
    assert result.output.count("nested_project") == 2


def test_app_multiple_lock_files_one_vulnerabilities_ignored_nested_pyproject_toml(
    temp_uv_lock_file: Path,
    temp_nested_uv_lock_file: Path,
    temp_pyproject_toml_file: Path,
    temp_nested_pyproject_toml_file_no_config: Path,
    no_vulnerabilities_response: HTTPXMock,
    one_vulnerability_response_v2: HTTPXMock,
) -> None:
    result = runner.invoke(
        app, [str(temp_uv_lock_file), str(temp_nested_uv_lock_file), "--disable-cache"]
    )
    assert result.exit_code == 2
    assert result.output.count("No vulnerabilities detected!") == 1
    assert result.output.count("Vulnerabilities detected!") == 1
